from enum import IntEnum
from typing import NamedTuple, List


class HSOpCode(IntEnum):
    GETFIELD = 0
    TEST = 1
    CALL_I = 2
    CALL_C = 3
    EQ = 4
    EQ_BK = 5
    GETGLOBAL = 6
    MOVE = 7
    SELF = 8
    RETURN = 9
    GETTABLE_S = 10
    GETTABLE_N = 11
    GETTABLE = 12
    LOADBOOL = 13
    TFORLOOP = 14
    SETFIELD = 15
    SETTABLE_S = 16
    SETTABLE_S_BK = 17
    SETTABLE_N = 18
    SETTABLE_N_BK = 19
    SETTABLE = 20
    SETTABLE_BK = 21
    TAILCALL_I = 22
    TAILCALL_C = 23
    TAILCALL_M = 24
    LOADK = 25
    LOADNIL = 26
    SETGLOBAL = 27
    JMP = 28
    CALL_M = 29
    CALL = 30
    INTRINSIC_INDEX = 31
    INTRINSIC_NEWINDEX = 32
    INTRINSIC_SELF = 33
    INTRINSIC_LITERAL = 34
    INTRINSIC_NEWINDEX_LITERAL = 35
    INTRINSIC_SELF_LITERAL = 36
    TAILCALL = 37
    GETUPVAL = 38
    SETUPVAL = 39
    ADD = 40
    ADD_BK = 41
    SUB = 42
    SUB_BK = 43
    MUL = 44
    MUL_BK = 45
    DIV = 46
    DIV_BK = 47
    MOD = 48
    MOD_BK = 49
    POW = 50
    POW_BK = 51
    NEWTABLE = 52
    UNM = 53
    NOT = 54
    LEN = 55
    LT = 56
    LT_BK = 57
    LE = 58
    LE_BK = 59
    CONCAT = 60
    TESTSET = 61
    FORPREP = 62
    FORLOOP = 63
    SETLIST = 64
    CLOSE = 65
    CLOSURE = 66
    VARARG = 67
    TAILCALL_I_R1 = 68
    CALL_I_R1 = 69
    SETUPVAL_R1 = 70
    TEST_R1 = 71
    NOT_R1 = 72
    GETFIELD_R1 = 73
    SETFIELD_R1 = 74
    NEWSTRUCT = 75
    DATA = 76
    SETSLOTN = 77
    SETSLOTI = 78
    SETSLOT = 79
    SETSLOTS = 80
    SETSLOTMT = 81
    CHECKTYPE = 82
    CHECKTYPES = 83
    GETSLOT = 84
    GETSLOTMT = 85
    SELFSLOT = 86
    SELFSLOTMT = 87
    GETFIELD_MM = 88
    CHECKTYPE_D = 89
    GETSLOT_D = 90
    GETGLOBAL_MEM = 91
    NUM_OPCODES = 92


class HSType(IntEnum):
    TNIL = 0
    TBOOLEAN = 1
    TLIGHTUSERDATA = 2
    TNUMBER = 3
    TSTRING = 4
    TTABLE = 5
    TFUNCTION = 6
    TUSERDATA = 7
    TTHREAD = 8
    TIFUNCTION = 9
    TCFUNCTION = 10
    TUI64 = 11
    TSTRUCT = 12


class HSOpArgMode(IntEnum):
    NUMBER = 0
    REG = 1
    CONST = 2


class HSOpArgModeA(IntEnum):
    UNUSED = 0
    REG = 1


class HSOpMode(IntEnum):
    ABC = 0
    ABX = 1
    ASBX = 2


class HSOpArgModeBC(IntEnum):
    UNUSED = 0
    NUMBER = 1
    OFFSET = 2
    REG = 3
    REG_OR_CONST = 4
    CONST = 5


class HSOpModes(NamedTuple):
    opCode: HSOpCode
    opMode: HSOpMode
    opArgModeA: HSOpArgModeA
    opArgModeB: HSOpArgModeBC
    opArgModeC: HSOpArgModeBC


HSOpModeList: List[HSOpModes] = [
    HSOpModes(HSOpCode.GETFIELD, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.CONST),
    HSOpModes(HSOpCode.TEST, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.UNUSED, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.CALL_I, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.CALL_C, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.EQ, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.EQ_BK, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.GETGLOBAL, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.MOVE, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.SELF, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.RETURN, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.GETTABLE_S, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.GETTABLE_N, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.GETTABLE, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.LOADBOOL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.TFORLOOP, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.UNUSED, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.SETFIELD, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE_S, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE_S_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE_N, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE_N_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETTABLE_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.TAILCALL_I, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.TAILCALL_C, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.TAILCALL_M, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.LOADK, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.LOADNIL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.SETGLOBAL, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.JMP, HSOpMode.ASBX, HSOpArgModeA.UNUSED, HSOpArgModeBC.OFFSET, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.CALL_M, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.CALL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_INDEX, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_NEWINDEX, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_SELF, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_LITERAL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_NEWINDEX_LITERAL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.INTRINSIC_SELF_LITERAL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.TAILCALL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.GETUPVAL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.SETUPVAL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.ADD, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.ADD_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SUB, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SUB_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.MUL, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.MUL_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.DIV, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.DIV_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.MOD, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.MOD_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.POW, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.POW_BK, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.NEWTABLE, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.UNM, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.NOT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.LEN, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.LT, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.LT_BK, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.LE, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.LE_BK, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.REG_OR_CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.CONCAT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.TESTSET, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.FORPREP, HSOpMode.ASBX, HSOpArgModeA.REG, HSOpArgModeBC.OFFSET, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.FORLOOP, HSOpMode.ASBX, HSOpArgModeA.REG, HSOpArgModeBC.OFFSET, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.SETLIST, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.OFFSET),
    HSOpModes(HSOpCode.CLOSE, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.UNUSED, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.CLOSURE, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.VARARG, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.TAILCALL_I_R1, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.CALL_I_R1, HSOpMode.ABC, HSOpArgModeA.UNUSED, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.SETUPVAL_R1, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.TEST_R1, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.UNUSED, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.NOT_R1, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.GETFIELD_R1, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.CONST),
    HSOpModes(HSOpCode.SETFIELD_R1, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.NEWSTRUCT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.DATA, HSOpMode.ABX, HSOpArgModeA.UNUSED, HSOpArgModeBC.OFFSET, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.SETSLOTN, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.UNUSED, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.SETSLOTI, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETSLOT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.SETSLOTS, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.REG),
    HSOpModes(HSOpCode.SETSLOTMT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.REG_OR_CONST),
    HSOpModes(HSOpCode.CHECKTYPE, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.CHECKTYPES, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.GETSLOT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.GETSLOTMT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.SELFSLOT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.SELFSLOTMT, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.GETFIELD_MM, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.CONST),
    HSOpModes(HSOpCode.CHECKTYPE_D, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.NUMBER, HSOpArgModeBC.UNUSED),
    HSOpModes(HSOpCode.GETSLOT_D, HSOpMode.ABC, HSOpArgModeA.REG, HSOpArgModeBC.REG, HSOpArgModeBC.NUMBER),
    HSOpModes(HSOpCode.GETGLOBAL_MEM, HSOpMode.ABX, HSOpArgModeA.REG, HSOpArgModeBC.CONST, HSOpArgModeBC.NUMBER),
]
